from mock import patch, Mock
from tests.helpers.testbase import TestBase
from cerami.request.search_attribute import SearchAttribute
from cerami.request.mixins import Limitable

class TestRequest(Limitable):
    pass

class TestLimitable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    def test_limit(self):
        """ it adds Limit"""
        self.request.add_attribute = Mock()
        self.request.limit(1)
        self.request.add_attribute.assert_called_with(
            SearchAttribute,
            'Limit',
            1)



from ..search_attribute import (
    DictAttribute,
    ProjectionExpressionAttribute)
from ...datatype.expression import BaseExpression

class Projectable(object):
    """A mixin to add the project method"""

    def project(self, *datatypes_or_expressions):
        """return a new Request setup with project attributes

        ProjectionExpression, ExpressionAttributeNames

        The args can be an array of datatypes or expressions because projecting 
        allows for list index expressions
        """
        for val in datatypes_or_expressions:
            names = {}
            if hasattr(val, 'expression_attribute_name'):
                expr = val
                names[expr.expression_attribute_name] = expr.datatype.column_name
            else:
                expr = BaseExpression('', val, None)
                names[expr.expression_attribute_name] = val.column_name
            self.add_attribute(
                ProjectionExpressionAttribute,
                'ProjectionExpression',
                expr)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
        return self



from .expression import (
    EqualityExpression,
    InExpression)

class DynamoDataType(object):
    """Abstract class for all DataTypes
    class variable defined. It is used for querying to
    determine what type of attribute is being searched upon
    """
    def __init__(
            self,
            default=None,
            column_name="",
            condition_type="",
            mapper_cls=None):
        self.default = default
        self.set_column_name(column_name)
        self.condition_type = condition_type
        self.mapper = mapper_cls(self) if mapper_cls else None

    def eq(self, value):
        return EqualityExpression("=", self, value)

    def neq(self, value):
        return EqualityExpression("<>", self, value)

    def gt(self, value):
        return EqualityExpression(">", self, value)

    def gte(self, value):
        return EqualityExpression(">=", self, value)

    def lt(self, value):
        return EqualityExpression("<", self, value)

    def lte(self, value):
        return EqualityExpression("<=", self, value)

    def in_(self, *values):
        return InExpression(self, values)

    def set_column_name(self, val):
        self.column_name = val

    def build(self, val):
        """set the value to val or default if present

        building is called automatically by the DynamoDataAttribute
        when the model is initialized.
        """
        if val is None:
            return self._get_default(val)
        return val

    def _get_default(self, val=None):
        if self.default:
            if callable(self.default):
                return self.default()
            else:
                return self.default
        return None



from copy import copy
from .base_datatype import DynamoDataType
from .mapper import ModelMapper

class ModelMap(DynamoDataType):
    def __init__(self, model_cls, mapper_cls=ModelMapper, **kwargs):
        super(ModelMap, self).__init__(condition_type="M", mapper_cls=mapper_cls, **kwargs)
        self.model_cls = model_cls
        for column in self.model_cls._columns:
            setattr(self, column.column_name, copy(column))

    def set_column_name(self, val):
        super(ModelMap, self).set_column_name(val)
        for name, attr in self.__dict__.items():
            if isinstance(attr, DynamoDataType):
                new_name = val + "." + name
                attr.set_column_name(new_name)



from ..search_attribute import SearchAttribute

class Returnable(object):
    def returns(self, value):
        self.add_attribute(
            SearchAttribute,
            'ReturnValues',
            value)
        return self



from .base_datatype import DynamoDataType
from .mapper import ByteMapper

class ByteBuffer(DynamoDataType):
    def __init__(self, mapper_cls=ByteMapper, **kwargs):
        super(ByteBuffer, self).__init__(mapper_cls=mapper_cls, condition_type="B", **kwargs)



class BaseDatatypeMapper(object):
    def __init__(self, datatype):
        self.datatype = datatype
        self.condition_type = self.datatype.condition_type

    def map(self, value):
        """return the value and its condition_type

        Mapping is done when converting the model into a form
        readable by DynamoDB. Mapping involves two steps.
        First it must return a dict of the value.
        Second it must "resolve" the value so DynamoDB can process it.
        ---
        ex) {"S": "Zac"}

        """
        if value == None:
            return {'NULL': True}
        res = {}
        res[self.condition_type] = self.resolve(value)
        return res

    def resolve(self, value):
        """returns the value resolved for dynamodb

        That is to say it is how the database expects
        it to be passed for all of the operations
        All classes that inherit from this class must implemenent
        a resolver.
        """
        return value

    def reconstruct(self, mapped_dict):
        """return the value from the mapped dict for model instantiation

        DynamoDB returns all attributes as a dict. Reconstructing
        reads this dict and "parses" the value. The return value
        can be used as the attribute on the Model.
        Reconstructing is only responsible for parsing the data
        as-is from DynamoDB. Default values are handled by the
        Datatype itself.
        """
        if mapped_dict.get('NULL') == True:
            return None
        return self.parse(mapped_dict[self.condition_type])

    def parse(self, value):
        """Convert the value from DynamoDB into a format for the Model

        This is the opposite of mapping. So when given
        {"S": "Zac"} it will return "Zac"
        value should never be None, reconstruct will not pass it here
        """
        return value



from .mixins import BaseRequest, Keyable, Projectable
from ..response import GetResponse

class GetRequest(BaseRequest, Keyable, Projectable):
    def execute(self):
        response = self.client.get_item(**self.build())
        return GetResponse(response, self.reconstructor)




from mock import Mock, patch, call
from tests.helpers.testbase import TestBase
from cerami.response import SaveResponse
from cerami.request import PutRequest
from cerami.request.search_attribute import DictAttribute

class TestPutRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = PutRequest(
            tablename="test",
            client=self.mocked_client)

    def test_execute(self):
        """it calls put_item with the build
        it returns a SaveResponse
        """
        self.request.build = Mock()
        expected = {"fake": True}
        self.mocked_client.put_item.return_value = {'Test': True}
        self.request.build.return_value = expected
        res = self.request.execute()
        self.mocked_client.put_item.assert_called_with(fake=True)
        assert isinstance(res, SaveResponse)

    def test_item(self):
        """it calls adds the Item attribute with the item_dict"""
        self.request.add_attribute = Mock()
        fake_data = {}
        self.request.item(fake_data)
        calls = [call(DictAttribute, 'Item', fake_data)]
        self.request.add_attribute.assert_has_calls(calls)



from .base_datatype_mapper import BaseDatatypeMapper

class UniformListMapper(BaseDatatypeMapper):
    def __init__(self, mapper):
        self.mapper = mapper
        self.condition_type = "L"

    def resolve(self, value):
        return [self.mapper.map(i) for i in value]

    def parse(self, value):
        return [self.mapper.reconstruct(i) for i in value]



from ..response import SearchResponse
from .mixins import BaseRequest, Filterable, Projectable, Limitable

class ScanRequest(
        BaseRequest,
        Filterable,
        Projectable,
        Limitable):
    def execute(self):
        response = self.client.scan(**self.build())
        return SearchResponse(response, self.reconstructor)



from mock import Mock, patch, call
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.response import SearchResponse
from cerami.request import QueryRequest
from cerami.request.search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute)
from cerami.request.mixins import Filterable, Keyable, Projectable, Limitable

class TestModel(Model):
    test = String()

class TestQueryRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = QueryRequest(
            tablename="test",
            client=self.mocked_client)

    def test_is_filterable(self):
        """it is filterable"""
        assert isinstance(self.request, Filterable)

    def test_is_keyable(self):
        """it is keyable"""
        assert isinstance(self.request, Keyable)

    def test_is_projectable(self):
        assert isinstance(self.request, Projectable)

    def test_is_limitable(self):
        assert isinstance(self.request, Limitable)

    def test_index(self):
        """it adds the IndexName to the request"""
        self.request.add_attribute = Mock()
        self.request.index('test-index')
        self.request.add_attribute.assert_called_with(
            SearchAttribute,
            'IndexName',
            'test-index')

    def test_execute(self):
        """it calls query with the build
        it returns a SearchResponse"""
        with patch("cerami.request.mixins.BaseRequest.build") as build:
            expected = {"fake": True}
            self.mocked_client.query.return_value = {
                'Count': 0,
                'ScannedCount': 0}
            build.return_value = expected
            res = self.request.execute()
            self.mocked_client.query.assert_called_with(fake=True)
            assert isinstance(res, SearchResponse)

    def test_key(self):
        self.request.add_attribute = Mock()
        expression = TestModel.test.eq('123')
        names = {}
        names[expression.expression_attribute_name] = 'test'

        self.request.key(expression)
        calls = [
            call(QueryExpressionAttribute,
                 'KeyConditionExpression',
                 expression),
            call(DictAttribute,
                 'ExpressionAttributeNames',
                 names),
            call(DictAttribute,
                 'ExpressionAttributeValues',
                 expression.value_dict())
        ]
        self.request.add_attribute.assert_has_calls(calls)



from .mixins import BaseRequest, Filterable, Keyable, Projectable, Limitable
from ..response import SearchResponse
from .search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute)

class QueryRequest(
        BaseRequest,
        Filterable,
        Keyable,
        Projectable,
        Limitable):
    def execute(self):
        response = self.client.query(**self.build())
        return SearchResponse(response, self.reconstructor)

    def index(self, index_name):
        """add IndexName to the request"""
        self.add_attribute(SearchAttribute, 'IndexName', index_name)
        return self

    def key(self, *expressions):
        """return a new SearchInterface setup with query attributes
        KeyConditionExpression, ExpressionAttributeNames,
        ExpressionAttributeValues are all required to query properly

        This completely overrides the implementation of Keyable
        but is defined as such for a simpler interface to remember.
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(
                QueryExpressionAttribute,
                'KeyConditionExpression',
                expression)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeValues',
                expression.value_dict())
        return self



from .base_datatype_mapper import BaseDatatypeMapper

class SetMapperDecorator(BaseDatatypeMapper):
    def __init__(self, mapper):
        self.mapper = mapper
        self.condition_type = self.mapper.datatype.condition_type + "S"

    def resolve(self, value):
        return [self.mapper.resolve(i) for i in value]

    def parse(self, value):
        return [self.mapper.parse(i) for i in value]



from .dynamic import List
from .mapper import UniformListMapper

class UniformList(List):
    def __init__(self, datatype, **kwargs):
        super(UniformList, self).__init__(**kwargs)
        self.datatype = datatype
        self.mapper = UniformListMapper(self.datatype.mapper)

    def index(self, idx):
        """return an expression attribute of the inner datatype
        sets the index value on the expression attribute 
        """
        datatype_cls = type(self.datatype)
        dt = datatype_cls(column_name=self.column_name)
        dt._index = idx
        return dt



import dateutil.parser
from datetime import datetime, timezone
from .base_string import BaseString
from .mapper import DatetimeMapper

class Datetime(BaseString):
    def __init__(self, mapper_cls=DatetimeMapper, **kwargs):
        super(Datetime, self).__init__(mapper_cls=mapper_cls, **kwargs)



from mock import Mock, patch
from datetime import datetime, timezone
from tests.helpers.testbase import TestBase
from cerami.datatype import Datetime
from cerami.datatype.mapper import DatetimeMapper

class TestDatetimeMapper(TestBase):
    def setUp(self):
        self.dt = Datetime()
        self.mapper = DatetimeMapper(self.dt)

    def test_resolve(self):
        """it adds the utc timezone and calls isoformat"""
        val = Mock()
        mocked_utc_datetime = Mock()
        val.replace.return_value = mocked_utc_datetime
        mocked_utc_datetime.isoformat.return_value = "fake isostring"
        res = self.mapper.resolve(val)
        val.replace.assert_called_with(tzinfo=timezone.utc)
        assert res == "fake isostring"

    def test_parse(self):
        """it uses dateutil.parser.parse"""
        with patch('dateutil.parser.parse') as parser:
            parser.return_value = 'parsed'
            assert self.mapper.parse('2020-01-01T00:00:00Z') == 'parsed'



from ..search_attribute import DictAttribute

class Keyable(object):
    def key(self, *expressions):
        """return a new Request setup with the Key attribute"""
        for expression in expressions:
            key_dict = {}
            key_dict[expression.datatype.column_name] = expression.attribute_map()
            self.add_attribute(DictAttribute, 'Key', key_dict)
        return self



from .string_mapper import StringMapper
from .base_datatype_mapper import BaseDatatypeMapper

class NumberMapper(BaseDatatypeMapper):
    def resolve(self, value):
        """convert the number into a string"""
        return str(value)

    def parse(self, value):
        """convert the value into an int"""
        return int(value)




from .base_request_test import TestBaseRequest
from .filterable_test import TestFilterable
from .keyable_test import TestKeyable
from .returnable_test import TestReturnable
from .projectable_test import TestProjectable
from .limitable_test import TestLimitable



from ..search_attribute import (
    DictAttribute,
    QueryExpressionAttribute)

class Filterable(object):
    """A mixin to add the filter method"""

    def filter(self, *expressions):
        """return a new Request setup with filter attributes

        FilterExpression, ExpressionAttributeNames, ExpressionAttributeValues
        are all required to filter properly
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(
                QueryExpressionAttribute,
                'FilterExpression',
                expression)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeValues',
                expression.value_dict())
        return self



from .base_datatype import DynamoDataType
from .mapper import StringMapper

class BaseString(DynamoDataType):
    def __init__(self, mapper_cls=StringMapper, **kwargs):
        super(BaseString, self).__init__(mapper_cls=mapper_cls, condition_type="S", **kwargs)



from .base_datatype_mapper import BaseDatatypeMapper

class ListMapper(BaseDatatypeMapper):
    def __init__(self, datatype, map_guesser, parse_guesser):
        super(ListMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def map(self, value):
        """map each v in the value list

        Use the MapGuesser to find the datatype
        Use the datatype's mapper to resolve the value
        """
        mapped = {}
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.map_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.map(val))
        mapped[self.datatype.condition_type] = res
        return mapped

    def resolve(self, value):
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.map_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.resolve(val))
        return res

    def parse(self, value):
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.parse_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.reconstruct(val))
        return res



class DynamoDataAttribute(object):
    def __init__(self, datatype, value=None, initialized=True):
        """initialize the DynamoDataAttribute

        This class stores the current value of the attribute on
        the model and some extra metadata about the attribute itself
        datatype - the Datatype of the corresponding Model attribute
        value - the current value
        initialized - if this attribute was included when the Model
            was instantiated
        _changed - has the set() method be called

        initialized / _changed help determine what attributes to include
        in the model.put() and model.update()

        value is built using datatype.build() to handle default values
        """
        self.datatype = datatype
        self.value = self.datatype.build(value)
        self.initialized = initialized
        self._changed = False

    def get(self):
        return self.value

    def set(self, newvalue):
        self._changed = True
        self.value = newvalue



from .base_datatype_mapper import BaseDatatypeMapper

class DictMapper(BaseDatatypeMapper):
    def __init__(self, datatype, map_guesser, parse_guesser):
        """initialize the DictMapper
        takes in an additional datatype_guesser
        so it has a way to assigning mapped values
        """
        super(DictMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def resolve(self, value):
        """map each k,v in the value dict

        Use the MapGuesser to find the datatype
        Use the datatype's mapper to resolve the value
        """
        res = {}
        for key, val in value.items():
            guessed_dt = self.map_guesser.guess(key, val)
            res[key] = guessed_dt.mapper.map(val)
        return res

    def parse(self, value):
        """map each k,v in the value dict

        Use the ParseGuesser to find the datatype
        User the datatype's reconstructor to parse the value
        """
        res = {}
        data = value[self.datatype.condition_type]
        for key, nested_dict in data.items():
            guessed_dt = self.parse_guesser.guess(key, nested_dict)
            res[key] = guessed_dt.mapper.reconstruct(nested_dict)
        return res



import dateutil.parser
from .base_datatype_mapper import BaseDatatypeMapper
from datetime import datetime, timezone

class DatetimeMapper(BaseDatatypeMapper):
    def resolve(self, value):
        """Convert the datetime into an ISO 8601 string

        All DynamoDB datetimes must be shifted to UTC
        https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/DynamoDBMapper.DataTypes.html
        """
        return value.replace(tzinfo=timezone.utc).isoformat()

    def parse(self, value):
        """Parse the datetime string

        Since it uses dateutil.parser, this should handle both
        ISO 8601 strings and timezone agnostic values, however
        if the record was creating using the resolve() method, a
        utc timezone should automatically be included
        """
        return dateutil.parser.parse(value)



from mock import Mock, patch, call
from tests.helpers.testbase import TestBase
from cerami.response import SaveResponse
from cerami.request import UpdateRequest
from cerami.request.mixins import Keyable, Returnable
from cerami.request.search_attribute import (
    UpdateExpressionAttribute,
    UpdateAction,
    DictAttribute)

class TestUpdateRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = UpdateRequest(
            tablename="test",
            client=self.mocked_client)

    def test_is_keyable(self):
        """it is keyable"""
        assert isinstance(self.request, Keyable)

    def test_is_returnable(self):
        """it is returnable"""
        assert isinstance(self.request, Returnable)

    def test_set(self):
        """it adds the SET attribute"""
        fake_expression = {'fake': True}
        self.request.update_expression = Mock()
        self.request.set(fake_expression)
        self.request.update_expression.assert_called_with('SET', fake_expression)



from .base_datatype import DynamoDataType
from .mapper import SetMapperDecorator

class Set(DynamoDataType):
    def __init__(self, datatype, **kwargs):
        super(Set, self).__init__(**kwargs)
        self.datatype = datatype
        self.mapper = SetMapperDecorator(self.datatype.mapper)
        self.condition_type = self.datatype.condition_type + "S"



from copy import copy
from ...reconstructor import RawReconstructor
from ..search_attribute import SearchAttribute

class BaseRequest(object):
    """The Base Class for all requests

    It provides the default constructor and methods for building
    requests.
    """

    def __init__(self, client, tablename="", request_attributes=None, reconstructor=None):
        """constructor for base request"""
        self.request_attributes = copy(request_attributes or {})
        self.client = client
        self.reconstructor = reconstructor or RawReconstructor()
        if tablename:
            self.add_attribute(SearchAttribute, 'TableName', tablename)

    def __str__(self):
        return self.build().__str__()

    def add_attribute(self, attr_class, name, value):
        """add a search attribute to a duplicated instance
        @param attr_class - a SearchAttribute class
        @param name - the name of the attribute
        @param value - the value that will be added to the SearchAttribute
        """
        request_attribute = self.request_attributes.get(name, attr_class(name))
        request_attribute.add(value)
        self.request_attributes[name] = request_attribute

    def build(self):
        """build the dict used by dynamodb"""
        return dict((k, v.build()) for k, v in self.request_attributes.items())




from .base_datatype import DynamoDataType
from .mapper import NumberMapper
from .expression import ArithmeticExpression

class BaseNumber(DynamoDataType):
    def __init__(self, mapper_cls=NumberMapper, **kwargs):
        super(BaseNumber, self).__init__(mapper_cls=mapper_cls, condition_type="N", **kwargs)

    def add(self, value):
        return ArithmeticExpression('+', self, value)

    def subtract(self, value):
        return ArithmeticExpression('-', self, value)



from .mixins import BaseRequest, Keyable, Returnable
from ..response import DeleteResponse

class DeleteRequest(BaseRequest, Keyable, Returnable):
    def execute(self):
        response = self.client.delete_item(**self.build())
        return DeleteResponse(response, self.reconstructor)



import numbers
from .number import Number
from .string import String
from .set import Set
from .base_datatype import DynamoDataType
from .expression import ListAppendExpression
from .mapper import (
    DictMapper,
    ListMapper)

class DefaultMapGuesser(object):
    def guess(self, key, value):
        """guess the datatype from the value
        Guessing for mapping means we are trying to figure out the datatype
        in order to convert it into a dict usable by dynamodb. So this
        guesser will makes its guess based on value directly
        """
        if isinstance(value, numbers.Number):
            return Number()
        elif isinstance(value, dict):
            return Map()
        elif isinstance(value, list):
            return List()
        else:
            return String()


class DefaultParseGuesser(object):
    def guess(self, key, value):
        """guess the datatype from the key within value
        we are guessing on something like {'M': {'test': {'S': 'hello'}}}
        where key is 'test' and value is {'S': 'hello'}
        this will fetch the inner key ('S') and guess from this value
        """
        attr_key = list(value)[0]
        if attr_key == "N":
            return Number()
        elif attr_key == "S":
            return String()
        elif attr_key == "SS":
            return Set(String())
        elif attr_key == "NS":
            return Set(Number())
        elif attr_key == "L":
            return List()
        elif attr_key == "M":
            return Map()
        else:
            return String()


class Map(DynamoDataType):
    def __init__(
            self,
            map_guesser=None,
            parse_guesser=None,
            **kwargs):
        super(Map, self).__init__(condition_type="M", **kwargs)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()
        self.mapper = DictMapper(self, self.map_guesser, self.parse_guesser)

    def key(self, datatype, key):
        column_name = self.column_name + "." + key
        return type(datatype)(column_name=column_name)


class List(DynamoDataType):
    def __init__(
            self,
            map_guesser=None,
            parse_guesser=None,
            **kwargs):
        super(List, self).__init__(condition_type="L", **kwargs)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()
        self.mapper = ListMapper(self, self.map_guesser, self.parse_guesser)

    def append(self, array):
        if not isinstance(array, list):
            array = [array]
        return ListAppendExpression(self, array)

    def index(self, idx, datatype_cls):
        """return an expression attribute of the inner datatype
        sets the index value on the expression attribute 
        """
        dt = datatype_cls(column_name=self.column_name)
        dt._index = idx
        return dt



from ..search_attribute import SearchAttribute

class Limitable(object):
    """A mixin to add the limit method"""

    def limit(self, limit_number):
        """return a new Request setup with the limit attribute"""
        self.add_attribute(
            SearchAttribute,
            'Limit',
            limit_number)
        return self
